#include <stdio.h>
#include <stdlib.h>

#include <jpeglib.h>
#include <setjmp.h>

#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spPluginP.h>

#include <sp/spDIBitmapP.h>

typedef struct sp_jpeg_error_mgr {
    struct jpeg_error_mgr pub;
    jmp_buf setjmp_buffer;
} *sp_jpeg_error_ptr;

typedef struct _spJpegPluginInstance
{
    char type[SP_MAX_LINE];
    int type_index;
    
    FILE *fp;
    volatile spBool read_flag;
    volatile spBool write_flag;

    struct sp_jpeg_error_mgr err;
    struct jpeg_compress_struct comp;
    struct jpeg_decompress_struct decomp;
    
    long offset;
    unsigned char *buf;

    short compression_quality;
} *spJpegPluginInstance;

void spJpegErrorExit(j_common_ptr jcptr)
{
    sp_jpeg_error_ptr err = (sp_jpeg_error_ptr)jcptr->err;

    (*jcptr->err->output_message)(jcptr);

    longjmp(err->setjmp_buffer, 1);

    return;
}

static spBool spIsSupportedByPluginJpeg(const char *filename)
{
    spBool supported = SP_FALSE;

    if (spGetBaseName(filename) == filename && filename[0] == '.') {
	if (strcaseeq(filename, ".jpg") || strcaseeq(filename, ".jpeg")) {
	    supported = SP_TRUE;
	}
    } else {
	volatile FILE *fp = NULL;
	volatile spBool created = SP_FALSE;
	struct jpeg_decompress_struct decomp;
	struct sp_jpeg_error_mgr err;

	decomp.err = jpeg_std_error(&err.pub);
	err.pub.error_exit = spJpegErrorExit;
	
	if (setjmp(err.setjmp_buffer)) {
	    if (created) {
		jpeg_destroy_decompress(&decomp);
	    }
	    if (fp != NULL) {
		spCloseFile((FILE *)fp);
	    }
	    return SP_FALSE;
	}
	
	jpeg_create_decompress(&decomp);
	created = SP_TRUE;

	if ((fp = spOpenFile(filename, "rb")) != NULL) {
	    jpeg_stdio_src(&decomp, (FILE *)fp);
	    if (jpeg_read_header(&decomp, TRUE) == JPEG_HEADER_OK) {
		supported = SP_TRUE;
	    }
	    jpeg_destroy_decompress(&decomp);
	    
	    spCloseFile((FILE *)fp);
	}
    }
    
    return supported;
}

static spBool spInitPluginJpeg(const char *lang)
{
    /*spSetDebugLevel(10);*/

    return SP_TRUE;
}

static spBool spFreePluginJpeg(void)
{

    return SP_TRUE;
}

static void *spInitPluginInstanceJpeg(const char *lang)
{
    spJpegPluginInstance pinstance;
    
    pinstance = xalloc(1, struct _spJpegPluginInstance);
    memset(pinstance, 0, sizeof(struct _spJpegPluginInstance));
    pinstance->type_index = 0;
    pinstance->compression_quality = 75;
    
    return (void *)pinstance;
}

static spBool spFreePluginInstanceJpeg(void *instance)
{
    xfree(instance);
    return SP_TRUE;
}

static spBool spSetPluginFileTypeJpeg(void *instance, int index)
{
    spJpegPluginInstance pinstance = (spJpegPluginInstance)instance;

    if (index == 0) {
	pinstance->type_index = index;
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static int spGetPluginFileTypeJpeg(void *instance)
{
    spJpegPluginInstance pinstance = (spJpegPluginInstance)instance;
    
    return pinstance->type_index;
}

static const char *spGetPluginBestSuffixJpeg(void *instance)
{
    return ".jpg";
}

static unsigned long spGetPluginDIBCapsJpeg(void *instance)
{
    return SP_DI_BITMAP_CAPS_DIRECT_ACCESS | SP_DI_BITMAP_CAPS_PIXEL_STRIDE_3
	| SP_DI_BITMAP_CAPS_READ | SP_DI_BITMAP_CAPS_WRITE
	| SP_DI_BITMAP_CAPS_UPSIDE_UP | SP_DI_BITMAP_CAPS_UPSIDE_DOWN 
	| SP_DI_BITMAP_CAPS_INVERSE_FORMAT | SP_DI_BITMAP_CAPS_GRAYSCALE;
}

static spBool spClosePluginJpeg(void *instance)
{
    spJpegPluginInstance pinstance = (spJpegPluginInstance)instance;

    if (pinstance->buf != NULL) {
	xfree(pinstance->buf);
	pinstance->buf = NULL;
    }
    
    if (pinstance->write_flag) {
	jpeg_destroy_compress(&pinstance->comp);
    } else if (pinstance->read_flag) {
	jpeg_destroy_decompress(&pinstance->decomp);
    }

    if (pinstance->fp != NULL) {
	/* close file */
	spCloseFile(pinstance->fp);
	pinstance->fp = NULL;
    }
    
    return SP_TRUE;
}

static spPluginError spOpenPluginJpeg(void *instance, const char *filename, const char *mode)
{
    const char *mode2;
    spJpegPluginInstance pinstance = (spJpegPluginInstance)instance;

    pinstance->read_flag = SP_FALSE;
    pinstance->write_flag = SP_FALSE;
    pinstance->offset = 0;

    if (mode[0] == 'w') {
	mode2 = "wb";
    } else {
	mode2 = "rb";
    }
    
    /* open file */
    if (strnone(filename) || (pinstance->fp = spOpenFile(filename, mode2)) == NULL) {
	spDebug(10, "spOpenPluginJpeg", "can't open file: %s\n", filename);
	return SP_PLUGIN_ERROR_OPEN;
    }
    spDebug(80, "spOpenPluginJpeg", "filename = %s, mode = %s\n", filename, mode);

    if (mode[0] == 'w') {
	pinstance->comp.err = jpeg_std_error(&pinstance->err.pub);
	pinstance->err.pub.error_exit = spJpegErrorExit;
	if (setjmp(pinstance->err.setjmp_buffer)) {
	    if (pinstance->write_flag) {
		jpeg_destroy_compress(&pinstance->comp);
	    }
	    spDebug(10, "spOpenPluginJpeg", "jpeg library error\n");
	    return SP_PLUGIN_ERROR_OPEN;
	}
	
	jpeg_create_compress(&pinstance->comp);
	pinstance->write_flag = SP_TRUE;
	
	jpeg_stdio_dest(&pinstance->comp, pinstance->fp);
    } else {
	pinstance->decomp.err = jpeg_std_error(&pinstance->err.pub);
	pinstance->err.pub.error_exit = spJpegErrorExit;
	if (setjmp(pinstance->err.setjmp_buffer)) {
	    if (pinstance->read_flag) {
		jpeg_destroy_decompress(&pinstance->decomp);
	    }
	    return SP_PLUGIN_ERROR_OPEN;
	}
	
	jpeg_create_decompress(&pinstance->decomp);
	pinstance->read_flag = SP_TRUE;
	
	jpeg_stdio_src(&pinstance->decomp, pinstance->fp);
	jpeg_read_header(&pinstance->decomp, TRUE);
    }

    spDebug(80, "spOpenPluginJpeg", "done: write_flag = %d\n", pinstance->write_flag);
    
    return SP_PLUGIN_ERROR_SUCCESS;
}

static spBool spSetPluginOtherInfoJpeg(void *instance, const char *id, char *data)
{
    spJpegPluginInstance pinstance = (spJpegPluginInstance)instance;
    
    if (streq(id, SP_PLUGIN_INFO_DIBITMAP_COMPRESSION_QUALITY)) {
	short *compression_quality = (short *)data;
	pinstance->compression_quality = *compression_quality;
	return SP_TRUE;
    }
    
    return SP_FALSE;
}

static spBool spGetPluginOtherInfoJpeg(void *instance, const char *id, char *data)
{
    spJpegPluginInstance pinstance = (spJpegPluginInstance)instance;
    
    if (streq(id, SP_PLUGIN_INFO_DIBITMAP_COMPRESSION_QUALITY)) {
	short *compression_quality = (short *)data;
	*compression_quality = pinstance->compression_quality;
	return SP_TRUE;
    }
    
    return SP_FALSE;
}

static spBool spInitJpegCompressStruct(spDIBitmapInfoInternal *dibinfo, struct jpeg_compress_struct *comp)
{
    spDebug(80, "spInitJpegCompressStruct", "bit_size = %ld, pixel_stride = %ld, pixel_format = %lx\n",
	    dibinfo->bit_size, dibinfo->pixel_stride, dibinfo->pixel_format);
    
    if (dibinfo->bit_size != 8) {
	spDebug(10, "spInitJpegCompressStruct", "bit_size is wrong\n");
	return SP_FALSE;
    }
    if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK) {
	if ((SP_DI_PIXEL_FORMAT_MASK & dibinfo->pixel_format) != SP_DI_PIXEL_FORMAT_GRAY
	    || dibinfo->pixel_stride != 1) {
	    spDebug(10, "spInitJpegCompressStruct", "wrong grayscale format\n");
	    return SP_FALSE;
	}
	comp->input_components = 1;
	comp->in_color_space = JCS_GRAYSCALE;
    } else {
	if ((SP_DI_PIXEL_FORMAT_MASK & dibinfo->pixel_format) != SP_DI_PIXEL_FORMAT_RGB
	    || dibinfo->pixel_stride != 3) {
	    spDebug(10, "spInitJpegCompressStruct", "wrong rgb format\n");
	    return SP_FALSE;
	}
	comp->input_components = 3;
	comp->in_color_space = JCS_RGB;
    }
    
    comp->image_width = dibinfo->width;
    comp->image_height =dibinfo->height;

    spDebug(100, "spInitJpegCompressStruct", "done\n");
    
    return SP_TRUE;
}

static spBool spSetPluginDIBInfoJpeg(void *instance, spDIBitmapInfoInternal *dibinfo)
{
    spJpegPluginInstance pinstance = (spJpegPluginInstance)instance;
    
    if (pinstance->write_flag == SP_FALSE) {
        if ((dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_YUV_MASK)
            || (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_SINGLE_RGB_MASK)
            || (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_FLOAT_MASK)
            || (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK)
            || (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)
            || (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK)
            || (dibinfo->bit_size != 8)
            ) {
            return SP_FALSE;
        }
        if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK) {
            if ((SP_DI_PIXEL_FORMAT_MASK & dibinfo->pixel_format) != SP_DI_PIXEL_FORMAT_GRAY
                || dibinfo->pixel_stride != 1) {
                spDebug(10, "spSetPluginDIBInfoJpeg", "wrong grayscale format\n");
                return SP_FALSE;
            }
	} else {
            if ((SP_DI_PIXEL_FORMAT_MASK & dibinfo->pixel_format) != SP_DI_PIXEL_FORMAT_RGB
                || dibinfo->pixel_stride != 3) {
                spDebug(10, "spSetPluginDIBInfoJpeg", "wrong rgb format\n");
                return SP_FALSE;
            }
        }
        
        return SP_TRUE;
    }

    if (spInitJpegCompressStruct(dibinfo, &pinstance->comp) == SP_FALSE) {
	return SP_FALSE;
    }
    
    return SP_TRUE;
}
    
static spBool spGetPluginDIBInfoJpeg(void *instance, spDIBitmapInfoInternal *dibinfo, unsigned long *strict_mask)
{
    spJpegPluginInstance pinstance = (spJpegPluginInstance)instance;

    spDebug(80, "spGetPluginDIBInfoJpeg", "in\n");
    
    if (((*strict_mask & SP_DI_BITMAP_STRICT_YUV_STATE)
	 && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_YUV_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_SINGLE_RGB_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_SINGLE_RGB_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_FLOAT_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_FLOAT_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_PREMULTIPLIED_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_ORDER)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_BIT_SIZE)
	    && dibinfo->bit_size != 8)) {
	spDebug(10, "spGetPluginDIBInfoJpeg", "not supported strict_mask: %lx\n", *strict_mask);
	return SP_FALSE;
    }
    
    if (pinstance->read_flag) {
	if (setjmp(pinstance->err.setjmp_buffer)) {
	    spDebug(10, "spGetPluginDIBInfoJpeg", "error\n");
	    return SP_FALSE;
	}
    
	jpeg_calc_output_dimensions(&pinstance->decomp);
    
	if (pinstance->decomp.out_color_components == 1) {
	    if (((*strict_mask & SP_DI_BITMAP_STRICT_GRAYSCALE_STATE)
		 && !(dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK))
		|| ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
		    && dibinfo->pixel_stride != 1)) {
		spDebug(10, "spGetPluginDIBInfoJpeg", "not supported strict_mask for grayscale: %lx\n", *strict_mask);
		return SP_FALSE;
	    }
	    dibinfo->pixel_stride = 1;
	    dibinfo->pixel_format = SP_DI_PIXEL_FORMAT_GRAY;
	} else if (pinstance->decomp.out_color_components == 3) {
	    if (((*strict_mask & SP_DI_BITMAP_STRICT_GRAYSCALE_STATE)
		 && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK))
		|| ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
		    && dibinfo->pixel_stride != 3)) {
		spDebug(10, "spGetPluginDIBInfoJpeg", "not supported strict_mask for RGB: %lx\n", *strict_mask);
		return SP_FALSE;
	    }
	    dibinfo->pixel_stride = 3;
	    dibinfo->pixel_format = SP_DI_PIXEL_FORMAT_RGB;
	}

	dibinfo->width = pinstance->decomp.output_width;
	dibinfo->height = pinstance->decomp.output_height;
	dibinfo->line_stride = dibinfo->width * dibinfo->pixel_stride;
	dibinfo->bit_size = 8;
	if (!(*strict_mask & SP_DI_BITMAP_STRICT_UPSIDE_DOWN_STATE)) {
	    dibinfo->upside_down = SP_FALSE;
	}

	*strict_mask |= SP_DI_BITMAP_STRICT_ALL;
    } else {
	if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK) {
	    if ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
		&& dibinfo->pixel_stride != 1) {
		spDebug(10, "spGetPluginDIBInfoJpeg", "not supported strict_mask for grayscale: %lx\n", *strict_mask);
		return SP_FALSE;
	    } else {
		dibinfo->pixel_stride = 1;
	    }
	    dibinfo->pixel_format = SP_DI_PIXEL_FORMAT_GRAY;
	} else {
	    if ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
		&& dibinfo->pixel_stride != 3) {
		spDebug(10, "spGetPluginDIBInfoJpeg", "not supported strict_mask for RGB: %lx\n", *strict_mask);
		return SP_FALSE;
	    } else {
		dibinfo->pixel_stride = 3;
	    }
	    dibinfo->pixel_format = SP_DI_PIXEL_FORMAT_RGB;
	}
	
	if ((*strict_mask & SP_DI_BITMAP_STRICT_LINE_STRIDE)
	    && dibinfo->line_stride != dibinfo->width * dibinfo->pixel_stride) {
	    spDebug(10, "spGetPluginDIBInfoPng", "not supported line_stride\n");
	    return SP_FALSE;
	} else {
	    dibinfo->line_stride = dibinfo->width * dibinfo->pixel_stride;
	}
	dibinfo->bit_size = 8;
	
	if (!(*strict_mask & SP_DI_BITMAP_STRICT_UPSIDE_DOWN_STATE)) {
	    dibinfo->upside_down = SP_FALSE;
	}
    }

    spDebug(80, "spGetPluginDIBInfoJpeg", "done\n");
    
    return SP_TRUE;
}

static spBool spReadPluginDIBDataJpeg(void *instance, spDIBitmapInfoInternal *dibinfo, unsigned char *data)
{
    int i;
    unsigned char *ptr;
    long ptr_stride;
    JSAMPROW row[1];
    volatile spBool started = SP_FALSE;
    spJpegPluginInstance pinstance = (spJpegPluginInstance)instance;
    
    spDebug(80, "spReadPluginDIBDataJpeg", "in\n");
    
    if (pinstance->read_flag == SP_FALSE) return SP_FALSE;
    
    if (setjmp(pinstance->err.setjmp_buffer)) {
	spDebug(10, "spReadPluginDIBDataJpeg", "error\n");
	if (started) {
	    jpeg_finish_decompress(&pinstance->decomp);
	}
	return SP_FALSE;
    }

    jpeg_start_decompress(&pinstance->decomp);
    started = SP_TRUE;

    if (dibinfo->upside_down) {
	ptr = data + dibinfo->line_stride * (dibinfo->height - 1);
	ptr_stride = -dibinfo->line_stride;
    } else {
	ptr = data;
	ptr_stride = dibinfo->line_stride;
    }

    for (i = 0; i < dibinfo->height; i++) {
	row[0] = ptr;
	jpeg_read_scanlines(&pinstance->decomp, row, 1);
	ptr += ptr_stride;
    }
    
    jpeg_finish_decompress(&pinstance->decomp);
    
    spDebug(80, "spReadPluginDIBDataJpeg", "done\n");
    
    return SP_TRUE;
}

static spBool spWritePluginDIBDataJpeg(void *instance, spDIBitmapInfoInternal *dibinfo, unsigned char *data)
{
    int i;
    unsigned char *ptr;
    long ptr_stride;
    JSAMPROW row[1];
    volatile spBool started = SP_FALSE;
    spJpegPluginInstance pinstance = (spJpegPluginInstance)instance;

    spDebug(80, "spWritePluginDIBDataJpeg", "in\n");
    
    if (pinstance->write_flag == SP_FALSE) return SP_FALSE;
    
    if (setjmp(pinstance->err.setjmp_buffer)) {
	spDebug(10, "spWritePluginDIBDataJpeg", "error\n");
	if (started) {
	    jpeg_finish_compress(&pinstance->comp);
	}
	return SP_FALSE;
    }
    
    jpeg_set_defaults(&pinstance->comp);
    jpeg_set_quality(&pinstance->comp, pinstance->compression_quality, TRUE);
    
    jpeg_start_compress(&pinstance->comp, TRUE);
    started = SP_TRUE;

    if (dibinfo->upside_down) {
	ptr = data + dibinfo->line_stride * (dibinfo->height - 1);
	ptr_stride = -dibinfo->line_stride;
    } else {
	ptr = data;
	ptr_stride = dibinfo->line_stride;
    }
    
    for (i = 0; i < dibinfo->height; i++) {
	row[0] = ptr;
	jpeg_write_scanlines(&pinstance->comp, row, 1);
	ptr += ptr_stride;
    }
    
    jpeg_finish_compress(&pinstance->comp);

    spDebug(80, "spWritePluginDIBDataJpeg", "done\n");
    
    return SP_TRUE;
}

spBool spReadPluginRGBAJpeg(void *instance, int *r, int *g, int *b, int *a)
{
    long x;
    unsigned char *ptr;
    volatile spBool started = SP_FALSE;
    spJpegPluginInstance pinstance = (spJpegPluginInstance)instance;

    spDebug(100, "spReadPluginRGBAJpeg", "in: offset = %ld\n", pinstance->offset);

    if (pinstance->read_flag == SP_FALSE) return SP_FALSE;

    if (setjmp(pinstance->err.setjmp_buffer)) {
	spDebug(10, "spReadPluginRGBAJpeg", "error\n");
	if (started) {
	    jpeg_finish_decompress(&pinstance->decomp);
	}
	return SP_FALSE;
    }
    
    if (pinstance->offset == 0) {
	jpeg_start_decompress(&pinstance->decomp);
	started = SP_TRUE;
	
	spDebug(80, "spReadPluginRGBAJpeg", "output_width = %d, output_height = %d, output_components = %d\n",
		pinstance->decomp.output_width, pinstance->decomp.output_height, pinstance->decomp.output_components);
	pinstance->buf = xalloc(pinstance->decomp.output_width * pinstance->decomp.output_components, unsigned char);
    } else {
	started = SP_TRUE;
    }

    x = pinstance->offset % pinstance->decomp.output_width;
    if (x == 0) {
	JSAMPROW row[1];
	row[0] = pinstance->buf;
	jpeg_read_scanlines(&pinstance->decomp, row, 1);
    }
    ptr = pinstance->buf + x * pinstance->decomp.output_components;
    *r = *ptr++;
    *g = *ptr++;
    *b = *ptr++;
    spDebug(100, "spReadPluginRGBAJpeg", "rbg = %x%x%x\n", *r, *g, *b);
    
    pinstance->offset++;

    if (pinstance->offset == pinstance->decomp.output_width * pinstance->decomp.output_height) {
	spDebug(80, "spReadPluginRGBAJpeg", "call jpeg_finish_decompress\n");
	jpeg_finish_decompress(&pinstance->decomp);
	spDebug(80, "spReadPluginRGBAJpeg", "jpeg_finish_decompress done\n");
    }
    
    spDebug(100, "spReadPluginRGBAJpeg", "done\n");
    
    return SP_TRUE;
}

spBool spWritePluginRGBAJpeg(void *instance, int r, int g, int b, int a)
{
    long x;
    unsigned char *ptr;
    volatile spBool started = SP_FALSE;
    spJpegPluginInstance pinstance = (spJpegPluginInstance)instance;

    spDebug(100, "spWritePluginRGBAJpeg", "in\n");

    if (pinstance->write_flag == SP_FALSE) return SP_FALSE;

    if (setjmp(pinstance->err.setjmp_buffer)) {
	spDebug(10, "spWritePluginRGBAJpeg", "error\n");
	if (started) {
	    jpeg_finish_compress(&pinstance->comp);
	}
	return SP_FALSE;
    }
    
    if (pinstance->offset == 0) {
	jpeg_set_defaults(&pinstance->comp);
	jpeg_set_quality(&pinstance->comp, pinstance->compression_quality, TRUE);
	
	jpeg_start_compress(&pinstance->comp, TRUE);
	started = SP_TRUE;
	
	pinstance->buf = xalloc(pinstance->comp.image_width * pinstance->comp.input_components, unsigned char);
    } else {
	started = SP_TRUE;
    }

    x = pinstance->offset % pinstance->comp.image_width;
    ptr = pinstance->buf + x * pinstance->comp.input_components;
    *ptr++ = r;
    *ptr++ = g;
    *ptr++ = b;
    
    if (x == pinstance->comp.image_width - 1) {
	JSAMPROW row[1];
	row[0] = pinstance->buf;
	jpeg_write_scanlines(&pinstance->comp, row, 1);

	if (pinstance->offset == pinstance->comp.image_width * pinstance->comp.image_height - 1) {
	    jpeg_finish_compress(&pinstance->comp);
	}
    }
    
    pinstance->offset++;
    
    spDebug(100, "spWritePluginRGBAJpeg", "done\n");
    
    return SP_TRUE;
}

static const char *sp_jpeg_file_type_list[] = {
    "jpeg",
    NULL,
};
static const char *sp_jpeg_file_desc_list[] = {
    "JPEG (libjpeg)",
    NULL,
};
static const char *sp_jpeg_file_filter_list[] = {
    "*.jpg",
    NULL,
};
    
static spDIBitmapPluginRec sp_dibitmap_plugin_jpeg = {
    NULL,
    NULL,

    SP_PLUGIN_DIBITMAP,
    "jpeg",
    1,
    SP_PLUGIN_PRIORITY_LOWER,
    SP_PLUGIN_CAPS_THREAD_SAFE,
    spInitPluginJpeg,
    spFreePluginJpeg,
    "JPEG I/O Plugin",
    "JPEG I/O Plugin Version 0.1\nCopyright(C) Hideki Banno\nThis software is based in part on the work of the Independent JPEG Group.",
    
    spInitPluginInstanceJpeg,
    spFreePluginInstanceJpeg,
    NULL,
    NULL,

    SP_PLUGIN_DEVICE_FILE,
    sp_jpeg_file_type_list,
    sp_jpeg_file_desc_list,
    sp_jpeg_file_filter_list,
    spIsSupportedByPluginJpeg,
    spSetPluginFileTypeJpeg,  
    spGetPluginFileTypeJpeg,
    0L,
    NULL,
    NULL,
    spGetPluginBestSuffixJpeg,
    
    spGetPluginDIBCapsJpeg,
	
    spOpenPluginJpeg,              
    spClosePluginJpeg,
	
    NULL,
    spSetPluginDIBInfoJpeg,
    NULL,
    spGetPluginDIBInfoJpeg,
    spSetPluginOtherInfoJpeg,   
    spGetPluginOtherInfoJpeg,
	
    spReadPluginDIBDataJpeg,          
    spWritePluginDIBDataJpeg,
	
    spReadPluginRGBAJpeg,
    spWritePluginRGBAJpeg,
};

#if defined(SP_DIBITMAP_PLUGIN_USE_PNG) && (defined(GTK) || defined(USE_MOTIF))
spBool spAddJpegDIBitmapPlugin(void)
{
    return spAddPluginRecList((spPluginRec *)&sp_dibitmap_plugin_jpeg, "jpeg");
}

spBool spRemoveJpegDIBitmapPlugin(void)
{
    return spRemovePluginRecList((spPluginRec *)&sp_dibitmap_plugin_jpeg);
}
#else
#include <sp/spPluginMain.h>

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_dibitmap_plugin_jpeg;
}
#endif
