#include <sp/spBaseLib.h>
#include <sp/spComponentLib.h>
#include <sp/spCanvasP.h>
#include <sp/spGraphicsP.h>
#include <sp/spDrawP.h>

#include <sp/spDIBitmapP.h>

spBool spCheckBasicDIBitmapStrictMaskWin(spDIBitmapInfoInternal *info, unsigned long strict_mask,
					 long *ppixel_stride, long *pline_stride)
{
    long pixel_stride;
    long line_stride;
    
    spDebug(100, "spCheckBasicDIBitmapStrictMaskWin", "info->bit_size = %ld, info->pixel_stride = %ld, info->pixel_format = %lx\n",
	    info->bit_size, info->pixel_stride, info->pixel_format);
    
    if (((strict_mask & SP_DI_BITMAP_STRICT_GRAYSCALE_STATE)
	 && (info->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK))
	|| ((strict_mask & SP_DI_BITMAP_STRICT_YUV_STATE)
	    && (info->pixel_format & SP_DI_PIXEL_FORMAT_YUV_MASK))
	|| ((strict_mask & SP_DI_BITMAP_STRICT_BIT_SIZE) && info->bit_size != 8)
	|| ((strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
	    && !(info->pixel_stride == 3 || info->pixel_stride == 4))
	|| ((strict_mask & SP_DI_BITMAP_STRICT_ALPHA_ORDER)
	    && (info->pixel_format & SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK))
	|| ((strict_mask & SP_DI_BITMAP_STRICT_PIXEL_ORDER)
	    && !(info->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK))) {
	spDebug(10, "spCheckBasicDIBitmapStrictMaskWin", "initial check failed\n");
	return SP_FALSE;
    }

    if (info->pixel_format & SP_DI_PIXEL_FORMAT_LAST_NULL_MASK) {
	pixel_stride = 4;
    } else {
	pixel_stride = 3;
    }
    
    line_stride = 4 * ((info->width * (pixel_stride * 8) + 31) / 32);
    if (strict_mask & SP_DI_BITMAP_STRICT_LINE_STRIDE) {
	if (line_stride != info->line_stride) {
	    spDebug(10, "spCheckBasicDIBitmapStrictMaskWin", "line stride is wrong\n");
	    return SP_FALSE;
	}
    }

    if (ppixel_stride != NULL) *ppixel_stride = pixel_stride;
    if (pline_stride != NULL) *pline_stride = line_stride;

    spDebug(100, "spCheckBasicDIBitmapStrictMaskWin", "done\n");
    
    return SP_TRUE;
}

spBool spUpdateDIBitmapInfoArch(spDIBitmapInfoInternal *info, unsigned long *strict_mask)
{
    long pixel_stride;
    long line_stride;

    if (spCheckBasicDIBitmapStrictMaskWin(info, *strict_mask, &pixel_stride, &line_stride) == SP_FALSE) {
	return SP_FALSE;
    }
    
    if (pixel_stride == 4) {
	info->pixel_format &= SP_DI_PIXEL_FORMAT_BGRA; /* '&' to keep alpha setting */
    } else {
	info->pixel_format = SP_DI_PIXEL_FORMAT_BGR;
    }
    info->bit_size = 8;
    info->pixel_stride = pixel_stride;
    info->line_stride = line_stride;
    
    *strict_mask |= (SP_DI_BITMAP_STRICT_ALL & ~(SP_DI_BITMAP_STRICT_UPSIDE_DOWN_STATE));
    if (pixel_stride == 4) {
	*strict_mask &= ~SP_DI_BITMAP_STRICT_ALPHA_STATE;
    }
    
    return SP_TRUE;
}

void *spCreateDIBitmapArch(spDIBitmapInfoInternal *info, unsigned long strict_mask)
{
    char *buf;
    BITMAPINFOHEADER *pbmih;
    long pixel_stride;
    long line_stride;
    long data_size;
    long total_size;

    spDebug(80, "spCreateDIBitmapArch", "width = %ld, height = %ld, pixel_format = %lx, upside_down = %d\n",
	    info->width, info->height, info->pixel_format, info->upside_down);

    if (spCheckBasicDIBitmapStrictMaskWin(info, strict_mask, &pixel_stride, &line_stride) == SP_FALSE) {
        spDebug(10, "spCreateDIBitmapArch", "spCheckBasicDIBitmapStrictMaskWin failed\n");
	return NULL;
    }
    
    data_size = line_stride * info->height;
    total_size = sizeof(BITMAPINFOHEADER) + data_size;
    spDebug(80, "spCreateDIBitmapArch", "data_size = %ld, total_size = %ld\n", data_size, total_size);
    
    if ((buf = xalloc(total_size, char)) != NULL) {
	memset(buf, 0, total_size);

	pbmih = (BITMAPINFOHEADER *)buf;
	pbmih->biSize = sizeof(BITMAPINFOHEADER);
	pbmih->biWidth = info->width;
	if (info->upside_down == SP_TRUE) {
	    pbmih->biHeight = info->height;
	} else {
	    pbmih->biHeight = -info->height;
	}
	pbmih->biPlanes = 1;
	pbmih->biBitCount = (WORD)pixel_stride * 8;
	pbmih->biCompression = BI_RGB;
	pbmih->biXPelsPerMeter = 3780; /* 96dpi */
	pbmih->biYPelsPerMeter = 3780; /* 96dpi */

	if (pixel_stride == 4) {
	    info->pixel_format &= SP_DI_PIXEL_FORMAT_BGRA;
	} else {
	    info->pixel_format = SP_DI_PIXEL_FORMAT_BGR;
	}
	info->bit_size = 8;
	info->pixel_stride = pixel_stride;
	info->line_stride = line_stride;
	info->data = (unsigned char *)(buf + pbmih->biSize);
    } else {
        spDebug(10, "spCreateDIBitmapArch", "cannot allocate memory\n");
    }
    
    spDebug(80, "spCreateDIBitmapArch", "done\n");
    
    return buf;
}

spBool spDestroyDIBitmapArch(void *bitmap)
{
    if (bitmap != NULL) {
	xfree(bitmap);
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spCopyDIBitmapToImageArch(spDIBitmapInfoInternal *info, void *bitmap, spComponent dest, int src_x, int src_y,
				 int width, int height, int dest_x, int dest_y)
{
    int ret;

    spDebug(100, "spCopyDIBitmapToImageArch", "in\n");
    
    spDebug(100, "spCopyDIBitmapToImageArch", "pixel_format = %lx, bit_size = %ld\n", info->pixel_format, info->bit_size);
    if ((info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)
	|| info->bit_size != 8) {
	return spCopyDIBitmapToImageWinGP(info, bitmap, dest, src_x, src_y, width, height, dest_x, dest_y);
    }
	
    ret = SetDIBitsToDevice(SpPrimitiveArch(dest).memdc,
			    dest_x, dest_y, width, height,
			    src_x, src_y, 0, height,
			    info->data,
			    (BITMAPINFO *)bitmap, DIB_RGB_COLORS);

    spDebug(100, "spCopyDIBitmapToImageArch", "ret = %d\n", ret);
    
    if (ret == 0) {
	return SP_FALSE;
    } else {
	return SP_TRUE;
    }
}

spBool spStretchDIBitmapToImageArch(spDIBitmapInfoInternal *info, void *bitmap, spComponent dest,
				    int src_x, int src_y, int src_width, int src_height,
				    int dest_x, int dest_y, int dest_width, int dest_height)
{
    int ret;
    int oldMode;
    POINT oldPt;
    BOOL setBrushRet;
	
    spDebug(100, "spStretchDIBitmapToImageArch", "in: pixel_format = %lx, bit_size = %ld, pixel_stride = %ld\n",
            info->pixel_format, info->bit_size, info->pixel_stride);
    
    if ((info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)
	|| info->bit_size != 8) {
	return spStretchDIBitmapToImageWinGP(info, bitmap, dest, src_x, src_y, src_width, src_height,
					     dest_x, dest_y, dest_width, dest_height);
    }
	
    oldMode = SetStretchBltMode(SpPrimitiveArch(dest).memdc, COLORONCOLOR/*HALFTONE*/);
    //setBrushRet = SetBrushOrgEx(SpPrimitiveArch(dest).memdc, 0, 0, &oldPt);
    
    spDebug(100, "spStretchDIBitmapToImageArch", "memdc = %ld, src = (%d, %d, %d, %d), dest = (%d, %d, %d, %d)\n",
            (long)SpPrimitiveArch(dest).memdc, src_x, src_y, src_width, src_height, dest_x, dest_y, dest_width, dest_height);
    
    ret = StretchDIBits(SpPrimitiveArch(dest).memdc,
			dest_x, dest_y, dest_width, dest_height,
			src_x, src_y, src_width, src_height,
			info->data,
			(BITMAPINFO *)bitmap, DIB_RGB_COLORS, SRCCOPY);
    spDebug(100, "spStretchDIBitmapToImageArch", "ret = %d\n", ret);

    //if (setBrushRet) SetBrushOrgEx(SpPrimitiveArch(dest).memdc, oldPt.x, oldPt.y, NULL);
    if (oldMode) SetStretchBltMode(SpPrimitiveArch(dest).memdc, oldMode);
    
    if (ret == GDI_ERROR) {
	return SP_FALSE;
    } else {
	return SP_TRUE;
    }
}

void *spCopyImageToDIBitmapArch(spComponent src, spDIBitmapInfoInternal *info, void *bitmap, int src_x, int src_y,
				int width, int height, int dest_x, int dest_y)
{
    HWND hwnd;
    HDC hdc;
    HDC memdc = NULL;
    HBITMAP hbitmap = NULL;
    HGDIOBJ hold = NULL;
    int ret = 0;
    
    spDebug(100, "spCopyImageToDIBitmapArch", "in: pixel_format = %lx, bit_size = %ld, pixel_stride = %ld\n",
            info->pixel_format, info->bit_size, info->pixel_stride);
    
    if ((info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)
	|| info->bit_size != 8) {
	return spCopyImageToDIBitmapWinGP(src, info, bitmap, src_x, src_y, width, height, dest_x, dest_y);
    }

    hwnd = SpPrimitiveArch(src).hwnd;
    
    hdc = GetDC(hwnd);
    if (src_x == 0 && src_y == 0 && dest_x == 0 && dest_y == 0
	&& width == info->width && height == info->height) {
        spDebug(100, "spCopyImageToDIBitmapArch", "case of same size\n");
	hbitmap = SpPrimitiveArch(src).hbitmap;
	ret = 1;
    } else {
	memdc = CreateCompatibleDC(hdc);
	hbitmap = CreateCompatibleBitmap(hdc, info->width, info->height);
	hold = SelectObject(memdc, hbitmap);
	if (dest_x != 0 || dest_y != 0) {
            spDebug(100, "spCopyImageToDIBitmapArch", "call SetDIBitsToDevice\n");
    
	    ret = SetDIBitsToDevice(memdc,
				    0, 0, info->width, info->height,
				    0, 0, 0, info->height,
				    info->data,
				    (BITMAPINFO *)bitmap, DIB_RGB_COLORS);
	} else {
	    ret = 1;
	}
	if (ret > 0) {
            spDebug(100, "spCopyImageToDIBitmapArch", "call BitBlt\n");
	    if (BitBlt(memdc, dest_x, dest_y,
		       MIN(SpComponentPart(src).client_width - src_x, width),
		       MIN(SpComponentPart(src).client_height - src_y, height),
		       SpPrimitiveArch(src).memdc, src_x, src_y, SRCCOPY)) {
		ret = 1;
	    } else {
		/* error */
		ret = 0;
	    }
	}
    }
    if (ret > 0) {
        spDebug(100, "spCopyImageToDIBitmapArch", "call GetDIBits\n");
	ret = GetDIBits(hdc, hbitmap, 0, height, info->data,
			(BITMAPINFO *)bitmap, DIB_RGB_COLORS);
    }
    ReleaseDC(hwnd, hdc);

    if (hbitmap != SpPrimitiveArch(src).hbitmap) {
	SelectObject(memdc, hold);
	DeleteObject(hbitmap);
	DeleteDC(memdc);
    }

    spDebug(100, "spCopyImageToDIBitmapArch", "done: ret = %d\n", ret);
    
    if (ret == 0) {
	return NULL;
    } else {
	return bitmap;
    }
}

spBool spReverseDIBitmapArch(spDIBitmapInfoInternal *info, void *dibdata)
{
    return spReverseDIBitmapDefault(info, dibdata);
}

spBool spFlipDIBitmapArch(spDIBitmapInfoInternal *info, void *dibdata, spBool lr_flag)
{
    return spFlipDIBitmapDefault(info, dibdata, lr_flag);
}
