#include <sp/spDefs.h>
#include <sp/spBase.h>

#include <sp/spTopLevelP.h>
#include <sp/spDrawP.h>
#include <sp/spDrawingPluginP.h>

static HPEN g_null_pen = NULL;
static HBRUSH g_null_brush = NULL;
static HFONT g_sys_font = NULL;

static HGLOBAL g_hDevMode = NULL;
static HGLOBAL g_hDevNames = NULL;


typedef struct _spPluginInstancePrinter
{
    HWND parent_hwnd;
    PRINTDLG printdlg;
    PAGESETUPDLG pagesetupdlg;
    DOCINFO docinfo;
    spBool mono_only;
    double xfactor, yfactor;

    double paper_width_mm;
    double paper_height_mm;
    double left_margin_mm;
    double right_margin_mm;
    double top_margin_mm;
    double bottom_margin_mm;

    int num_page;
    int start_page;
    int end_page;
} *spPluginInstancePrinter;

static spBool spInitPrinter(const char *lang)
{
    if (g_null_pen == NULL) {
	g_null_pen = GetStockObject(NULL_PEN);
	g_null_brush = GetStockObject(HOLLOW_BRUSH);
	g_sys_font = GetStockObject(DEFAULT_GUI_FONT);
    }
    
    return SP_TRUE;
}

static spBool spFreePrinter(void)
{
    if (g_hDevMode != NULL) {
	GlobalFree(g_hDevMode);
    }
    g_hDevMode = NULL;
    if (g_hDevNames != NULL) {
	GlobalFree(g_hDevNames);
    }
    g_hDevNames = NULL;
  
    return SP_TRUE;
}

static void *spInitInstancePrinter(const char *lang)
{
    spPluginInstancePrinter pinstance;

    pinstance = xalloc(1, struct _spPluginInstancePrinter);
    memset(pinstance, 0, sizeof(struct _spPluginInstancePrinter));
    pinstance->printdlg.hDC = NULL;
    pinstance->printdlg.hDevMode = NULL;
    pinstance->printdlg.hDevNames = NULL;
    pinstance->xfactor = 1.0;
    pinstance->yfactor = 1.0;
    pinstance->paper_width_mm = 0.0;
    pinstance->paper_height_mm = 0.0;
    spGetPaperDimensions(SP_PAPER_SIZE_A4, SP_PAPER_ORIENTATION_PORTRAIT,
			 &pinstance->paper_width_mm, &pinstance->paper_height_mm);
    pinstance->left_margin_mm = 20.0;
    pinstance->right_margin_mm = 20.0;
    pinstance->top_margin_mm = 20.0;
    pinstance->bottom_margin_mm = 20.0;
    
    pinstance->num_page = 1;
    pinstance->start_page = 0;
    pinstance->end_page = 0;
    
    return (void *)pinstance;
}

static void resetPrintDlg(PRINTDLG *printdlg)
{
    if (printdlg->hDC != NULL) {
	DeleteDC(printdlg->hDC);
    }
    printdlg->hDC = NULL;

    return;
}

static spBool spFreeInstancePrinter(void *instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    resetPrintDlg(&pinstance->printdlg);
    
    xfree(instance);
    return SP_TRUE;
}

static spBool spCreateComponentPrinter(void *instance, void *parent_arch, spArg *args, int num_arg)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    if (parent_arch != NULL) {
	pinstance->parent_hwnd = (HWND)parent_arch;
    }
    
    return SP_TRUE;
}

static spBool spDestroyComponentPrinter(void *instance)
{
    return SP_TRUE;
}

static spBool spOpenSetupDialogPrinter(void *instance, spArg *args, int num_arg)
{
    BOOL ok;
    HWND hwnd;
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    spDebug(50, "spOpenSetupDialogPrinter",
	    "input: paper_width_mm = %f, paper_height_mm = %f, left_margin_mm = %f, right_margin_mm = %f, top_margin_mm = %f, bottom_margin_mm = %f\n",
	    pinstance->paper_width_mm, pinstance->paper_height_mm, pinstance->left_margin_mm, pinstance->right_margin_mm,
	    pinstance->top_margin_mm, pinstance->bottom_margin_mm);

    if (pinstance->parent_hwnd != NULL) {
	hwnd = pinstance->parent_hwnd;
    } else {
	hwnd = GetForegroundWindow();
    }
    
    pinstance->pagesetupdlg.lStructSize = sizeof(PAGESETUPDLG);
    pinstance->pagesetupdlg.hwndOwner = hwnd;
    pinstance->pagesetupdlg.hDevMode = g_hDevMode;
    pinstance->pagesetupdlg.hDevNames = g_hDevNames;
    pinstance->pagesetupdlg.Flags = PSD_INHUNDREDTHSOFMILLIMETERS | PSD_MARGINS;
    pinstance->pagesetupdlg.rtMargin.left = (LONG)spRound(pinstance->left_margin_mm * 100.0);
    pinstance->pagesetupdlg.rtMargin.right = (LONG)spRound(pinstance->right_margin_mm * 100.0);
    pinstance->pagesetupdlg.rtMargin.top = (LONG)spRound(pinstance->top_margin_mm * 100.0);
    pinstance->pagesetupdlg.rtMargin.bottom = (LONG)spRound(pinstance->bottom_margin_mm * 100.0);

    ok = PageSetupDlg(&pinstance->pagesetupdlg);
    
    if (pinstance->pagesetupdlg.hDevMode != NULL && g_hDevMode == NULL) {
	g_hDevMode = pinstance->pagesetupdlg.hDevMode;
    }
    if (pinstance->pagesetupdlg.hDevNames != NULL && g_hDevNames == NULL) {
	g_hDevNames = pinstance->pagesetupdlg.hDevNames;
    }
    
    if (ok) {
	pinstance->paper_width_mm = (double)pinstance->pagesetupdlg.ptPaperSize.x / 100.0;;
	pinstance->paper_height_mm = (double)pinstance->pagesetupdlg.ptPaperSize.y / 100.0;;
	pinstance->left_margin_mm = (double)pinstance->pagesetupdlg.rtMargin.left / 100.0;
	pinstance->right_margin_mm = (double)pinstance->pagesetupdlg.rtMargin.right / 100.0;
	pinstance->top_margin_mm = (double)pinstance->pagesetupdlg.rtMargin.top / 100.0;
	pinstance->bottom_margin_mm = (double)pinstance->pagesetupdlg.rtMargin.bottom / 100.0;
	spDebug(50, "spOpenSetupDialogPrinter",
		"output: paper_width_mm = %f, paper_height_mm = %f, left_margin_mm = %f, right_margin_mm = %f, top_margin_mm = %f, bottom_margin_mm = %f\n",
		pinstance->paper_width_mm, pinstance->paper_height_mm, pinstance->left_margin_mm, pinstance->right_margin_mm,
		pinstance->top_margin_mm, pinstance->bottom_margin_mm);
	
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static spBool spOpenTargetDialogPrinter(void *instance, spArg *args, int num_arg)
{
    int i;
    BOOL ok;
    HWND hwnd;
    DEVMODE *pdevmode;
    
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    if (pinstance->parent_hwnd != NULL) {
	hwnd = pinstance->parent_hwnd;
    } else {
	hwnd = GetForegroundWindow();
    }
    
    resetPrintDlg(&pinstance->printdlg);
    pinstance->printdlg.lStructSize = sizeof(PRINTDLG);
    pinstance->printdlg.hwndOwner = hwnd;
    pinstance->printdlg.hDevMode = g_hDevMode;
    pinstance->printdlg.hDevNames = g_hDevNames;
    pinstance->printdlg.Flags = PD_RETURNDC | PD_USEDEVMODECOPIESANDCOLLATE | PD_NOPAGENUMS | PD_NOSELECTION;
    pinstance->printdlg.nCopies = 1;
    pinstance->printdlg.nMinPage = 1;
    pinstance->printdlg.nMaxPage = pinstance->num_page;
    
    if (pinstance->start_page == 1 && pinstance->end_page == pinstance->num_page) {
        pinstance->printdlg.nFromPage = 0;
        pinstance->printdlg.nToPage = 0;
    } else {
        pinstance->printdlg.Flags &= ~PD_NOSELECTION;
        pinstance->printdlg.nFromPage = pinstance->start_page;
        pinstance->printdlg.nToPage = pinstance->end_page;
    }

    for (i = 0; i < num_arg; i++) {
	if (spEqParam(args[i].param, SppEnablePageRange)) {
	    if ((spBool)args[i].value == SP_TRUE) {
		pinstance->printdlg.Flags &= ~PD_NOPAGENUMS;
	    }
	} else if (spEqParam(args[i].param, SppEnablePageRangeSelection)) {
	    if ((spBool)args[i].value == SP_TRUE) {
		pinstance->printdlg.Flags &= ~PD_NOSELECTION;
	    }
	} else if (spEqParam(args[i].param, SppMaxPageRange)) {
	    pinstance->printdlg.nMaxPage = (int)args[i].value;
            pinstance->num_page = MAX(pinstance->num_page, pinstance->printdlg.nMaxPage);
	}
    }

    ok = PrintDlg(&pinstance->printdlg);

    if (pinstance->printdlg.hDevMode != NULL && g_hDevMode == NULL) {
	g_hDevMode = pinstance->printdlg.hDevMode;
    }
    if (pinstance->printdlg.hDevNames != NULL && g_hDevNames == NULL) {
	g_hDevNames = pinstance->printdlg.hDevNames;
    }
    
    if (ok) {
        if (pinstance->printdlg.Flags & PD_SELECTION) {
            pinstance->start_page = SP_PAGE_RANGE_SELECTION;
            pinstance->end_page = SP_PAGE_RANGE_SELECTION;
        } else if (pinstance->printdlg.Flags & PD_PAGENUMS) {
            pinstance->start_page = pinstance->printdlg.nFromPage;
            pinstance->end_page = pinstance->printdlg.nToPage;
        } else {
            pinstance->start_page = 1;
            pinstance->end_page = pinstance->num_page;
        }
        
	if ((pdevmode = (DEVMODE *)GlobalLock(pinstance->printdlg.hDevMode)) != NULL) {
	    if (pdevmode->dmColor == DMCOLOR_MONOCHROME) {
		pinstance->mono_only = SP_TRUE;
	    }
	    GlobalUnlock(pinstance->printdlg.hDevMode);
	}
	
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static spBool spCloseTargetPrinter(void *instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    resetPrintDlg(&pinstance->printdlg);
    
    return SP_TRUE;
}

spBool spSetPageInfoPrinter(void *instance, spPaperSize paper_size, spPaperOrientation paper_orientation,
	double left_margin_mm, double right_margin_mm, double top_margin_mm, double bottom_margin_mm)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    pinstance->left_margin_mm = left_margin_mm;
    pinstance->right_margin_mm = right_margin_mm;
    pinstance->top_margin_mm = top_margin_mm;
    pinstance->bottom_margin_mm = bottom_margin_mm;
    
    return spGetPaperDimensions(paper_size, paper_orientation,
				&pinstance->paper_width_mm, &pinstance->paper_height_mm);
}


spBool spGetPageInfoPrinter(void *instance, spPaperSize *paper_size, spPaperOrientation *paper_orientation,
        double *left_margin_mm, double *right_margin_mm, double *top_margin_mm, double *bottom_margin_mm)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    *left_margin_mm = pinstance->left_margin_mm;
    *right_margin_mm = pinstance->right_margin_mm;
    *top_margin_mm = pinstance->top_margin_mm;
    *bottom_margin_mm = pinstance->bottom_margin_mm;
    
    return spGetPaperFromDimensions(pinstance->paper_width_mm, pinstance->paper_height_mm,
				    paper_size, paper_orientation);
}

spBool spSetPageRangePrinter(void *instance, int start_page, int end_page, int max_page)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    pinstance->start_page = start_page;
    pinstance->end_page = end_page;
    pinstance->num_page = MAX(max_page, 1);
    spDebug(50, "spSetPageRangePrinter", "start_page = %d, end_page = %d, num_page = %d\n",
            pinstance->start_page, pinstance->end_page, pinstance->num_page);

    return SP_TRUE;
}
	
spBool spGetPageRangePrinter(void *instance, int *start_page, int *end_page)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    spDebug(50, "spGetPageRangePrinter", "start_page = %d, end_page = %d, num_page = %d\n",
            pinstance->start_page, pinstance->end_page, pinstance->num_page);
    
    *start_page = pinstance->start_page;
    *end_page = pinstance->end_page;

    return SP_TRUE;
}
	
unsigned long spGetDrawingCapsPrinter(void *instance)
{
    unsigned long caps = 0L;
    int wcaps;
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    caps |= SP_DRAWING_CAPS_STRING_EXTENT | SP_DRAWING_CAPS_STRING_EXTENT_ACCURATE
	| SP_DRAWING_CAPS_SCALING | SP_DRAWING_CAPS_MULTIPLE_PAGES;
    
    wcaps = GetDeviceCaps(pinstance->printdlg.hDC, RASTERCAPS);
    if (wcaps & RC_SCALING) caps |= SP_DRAWING_CAPS_SCALING;
    if (wcaps & RC_BITBLT) caps |= SP_DRAWING_CAPS_COPY_IMAGE;
    if (wcaps & RC_STRETCHBLT) caps |= SP_DRAWING_CAPS_COPY_IMAGE_STRETCHABLE;

    wcaps = GetDeviceCaps(pinstance->printdlg.hDC, LINECAPS);
    if (!(wcaps & LC_NONE)) {
	spDebug(50, "spGetDrawingCapsPrinter", "line supported\n");
	caps |= SP_DRAWING_CAPS_DRAW_LINE;
	if (wcaps & LC_WIDE) caps |= SP_DRAWING_CAPS_DRAW_WIDE_LINE;
	if (wcaps & LC_STYLED) caps |= SP_DRAWING_CAPS_DRAW_STYLED_LINE;
	if (spIsOSWinNT() == SP_TRUE) {
	    if (wcaps & LC_WIDESTYLED) caps |= SP_DRAWING_CAPS_DRAW_WIDE_STYLED_LINE;
	}
    }
    
    wcaps = GetDeviceCaps(pinstance->printdlg.hDC, CURVECAPS);
    if (wcaps & CC_PIE) {
	spDebug(50, "spGetDrawingCapsPrinter", "arc supported\n");
	caps |= SP_DRAWING_CAPS_DRAW_ARC;
	if (wcaps & CC_INTERIORS) caps |= SP_DRAWING_CAPS_FILL_ARC;
	if (wcaps & CC_WIDE) caps |= SP_DRAWING_CAPS_DRAW_WIDE_ARC;
	if (wcaps & CC_STYLED) caps |= SP_DRAWING_CAPS_DRAW_STYLED_ARC;
	if (spIsOSWinNT() == SP_TRUE) {
	    if (wcaps & CC_WIDESTYLED) caps |= SP_DRAWING_CAPS_DRAW_WIDE_STYLED_ARC;
	}
     }
    
    wcaps = GetDeviceCaps(pinstance->printdlg.hDC, POLYGONALCAPS);
    if (wcaps & PC_RECTANGLE) {
	spDebug(50, "spGetDrawingCapsPrinter", "rectangle supported\n");
	caps |= SP_DRAWING_CAPS_DRAW_RECTANGLE;
	if (wcaps & PC_INTERIORS) caps |= SP_DRAWING_CAPS_FILL_RECTANGLE;
	if (wcaps & PC_WIDE) caps |= SP_DRAWING_CAPS_DRAW_WIDE_RECTANGLE;
	if (wcaps & PC_STYLED) caps |= SP_DRAWING_CAPS_DRAW_STYLED_RECTANGLE;
	if (spIsOSWinNT() == SP_TRUE) {
	    if (wcaps & PC_WIDESTYLED) caps |= SP_DRAWING_CAPS_DRAW_WIDE_STYLED_RECTANGLE;
	}
    }
    if (wcaps & PC_POLYGON) {
	spDebug(50, "spGetDrawingCapsPrinter", "polygon supported\n");
	caps |= SP_DRAWING_CAPS_DRAW_POLYGON;
	if (wcaps & PC_INTERIORS) caps |= SP_DRAWING_CAPS_FILL_POLYGON;
	if (wcaps & PC_WIDE) caps |= SP_DRAWING_CAPS_DRAW_WIDE_POLYGON;
	if (wcaps & PC_STYLED) caps |= SP_DRAWING_CAPS_DRAW_STYLED_POLYGON;
	if (spIsOSWinNT() == SP_TRUE) {
	    if (wcaps & PC_WIDESTYLED) caps |= SP_DRAWING_CAPS_DRAW_WIDE_STYLED_POLYGON;
	}
    }

    wcaps = GetDeviceCaps(pinstance->printdlg.hDC, TEXTCAPS);
    spDebug(50, "spGetDrawingCapsPrinter", "wcaps = %lx, TC_CR_ANY = %ld, TC_CR_90 = %ld\n",
	    wcaps, !!(wcaps & TC_CR_ANY), !!(wcaps & TC_CR_90));
    if (wcaps & TC_CR_ANY) {
	caps |= SP_DRAWING_CAPS_STRING_ROTATION;
    }
	
    
    if (pinstance->mono_only == SP_TRUE) caps |= SP_DRAWING_CAPS_MONO_ONLY;

    return caps;
}

unsigned long spGetGraphicsModeCapsPrinter(void *instance)
{
    return 0L;
}

static spBool spIsDrawablePrinter(void *instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    if (pinstance->printdlg.hDC != NULL) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static void convertSizeMMToDot(HDC hDC, double paper_width_mm, double paper_height_mm,
			       double left_margin_mm, double right_margin_mm, double top_margin_mm, double bottom_margin_mm,
			       int *left_margin_dot, int *top_margin_dot, int *client_width_dot, int *client_height_dot)
{
    int x_offset_dot, y_offset_dot;
    double x_dp_mm, y_dp_mm;
    
    x_offset_dot = GetDeviceCaps(hDC, PHYSICALOFFSETX);
    y_offset_dot = GetDeviceCaps(hDC, PHYSICALOFFSETY);
    spDebug(50, "convertSizeMMToDot", "x_offset_dot = %d, y_offset_dot = %d\n",
	    x_offset_dot, y_offset_dot);

    x_dp_mm = (double)GetDeviceCaps(hDC, LOGPIXELSX) / 25.4;
    y_dp_mm = (double)GetDeviceCaps(hDC, LOGPIXELSY) / 25.4;
    spDebug(50, "convertSizeMMToDot", "x_dp_mm = %f, y_dp_mm = %f\n", x_dp_mm, y_dp_mm);

    if (client_width_dot != NULL) {
	*client_width_dot = (int)spRound((paper_width_mm - left_margin_mm - right_margin_mm) * x_dp_mm);
	spDebug(50, "convertSizeMMToDot", "client_width_dot = %d\n", *client_width_dot);
    }
    if (client_height_dot != NULL) {
	*client_height_dot = (int)spRound((paper_height_mm - top_margin_mm - bottom_margin_mm) * y_dp_mm);
	spDebug(50, "convertSizeMMToDot", "client_height_dot = %d\n", *client_height_dot);
    }
    
    if (left_margin_dot != NULL) {
	*left_margin_dot = MAX((int)spRound(left_margin_mm * x_dp_mm) - x_offset_dot, 0);
	spDebug(50, "convertSizeMMToDot", "left_margin_dot = %d\n", *left_margin_dot);
    }
    if (top_margin_dot != NULL) {
	*top_margin_dot = MAX((int)spRound(top_margin_mm * y_dp_mm) - y_offset_dot, 0);
	spDebug(50, "convertSizeMMToDot", "top_margin_dot = %d\n", *top_margin_dot);
    }

    return;
}

static spBool spGetClientSizePrinter(void *instance, int *width, int *height)
{
    int width_dot, height_dot;
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    convertSizeMMToDot(pinstance->printdlg.hDC, pinstance->paper_width_mm, pinstance->paper_height_mm,
		       pinstance->left_margin_mm, pinstance->right_margin_mm,
		       pinstance->top_margin_mm, pinstance->bottom_margin_mm,
		       NULL, NULL, &width_dot, &height_dot);
    spDebug(50, "spGetClientSizePrinter", "width_dot = %d, height_dot = %d\n", width_dot, height_dot);
    *width = (int)((double)width_dot / pinstance->xfactor);
    *height = (int)((double)height_dot / pinstance->yfactor);
    spDebug(50, "spGetClientSizePrinter", "width = %d, height = %d\n", *width, *height);

    return SP_TRUE;
}

static spBool spGetPhysicalSizePrinter(void *instance, double *width_mm, double *height_mm)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    *width_mm = pinstance->paper_width_mm - pinstance->left_margin_mm - pinstance->right_margin_mm;
    *height_mm = pinstance->paper_height_mm - pinstance->top_margin_mm - pinstance->bottom_margin_mm;
    
    return SP_TRUE;
}

static spBool spSetScalingFactorPrinter(void *instance, double xfactor, double yfactor)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    pinstance->xfactor = xfactor;
    pinstance->yfactor = yfactor;
    
    return SP_TRUE;
}
     
static spBool transformCoordinate(spPluginInstancePrinter pinstance, double xfactor, double yfactor)
{
    int left_margin_dot, top_margin_dot;
    int width_dot, height_dot;
    int dev_width, dev_height;
    POINT org_point;
    SIZE window_size;
    SIZE viewport_size;
    
    SetWindowOrgEx(pinstance->printdlg.hDC, 0, 0, NULL);
    
    /* move (left_margin_mm, top_margin_mm) to (0, 0) */
    left_margin_dot = top_margin_dot = 0;
    convertSizeMMToDot(pinstance->printdlg.hDC, pinstance->paper_width_mm, pinstance->paper_height_mm,
		       pinstance->left_margin_mm, pinstance->right_margin_mm,
		       pinstance->top_margin_mm, pinstance->bottom_margin_mm,
		       &left_margin_dot, &top_margin_dot, &width_dot, &height_dot);
    spDebug(50, "transformCoordinate", "left_margin = %f-->%d, top_margin = %f-->%d\n",
	    pinstance->left_margin_mm, left_margin_dot, pinstance->top_margin_mm, top_margin_dot);
    SetViewportOrgEx(pinstance->printdlg.hDC, left_margin_dot, top_margin_dot, &org_point);

    dev_width = GetDeviceCaps(pinstance->printdlg.hDC, HORZRES);
    dev_height = GetDeviceCaps(pinstance->printdlg.hDC, VERTRES);
    spDebug(50, "transformCoordinate", "dev_width = %d (%d), dev_height = %d (%d)\n",
	    dev_width, width_dot, dev_height, height_dot);

    if (width_dot > 0 && height_dot > 0
	&& SetMapMode(pinstance->printdlg.hDC, MM_ANISOTROPIC)) {
	if (SetWindowExtEx(pinstance->printdlg.hDC,
			   (int)spRound((double)width_dot / xfactor),
			   (int)spRound((double)height_dot / yfactor), &window_size)) {
	    spDebug(50, "transformCoordinate", "win width = %ld, win height = %ld\n",
		    window_size.cx, window_size.cy);
	    if (SetViewportExtEx(pinstance->printdlg.hDC,
				 width_dot, height_dot, &viewport_size)) {
		spDebug(50, "transformCoordinate", "viewport width = %ld, viewport height = %ld\n",
			viewport_size.cx, viewport_size.cy);
		return SP_TRUE;
	    }
	}
    }
    
    return SP_FALSE;
}
     
static spBool spStartDrawingPrinter(void *instance, const char *job_name)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    transformCoordinate(pinstance, pinstance->xfactor, pinstance->yfactor);
    
    memset(&pinstance->docinfo, 0, sizeof(DOCINFO));
    pinstance->docinfo.cbSize = sizeof(DOCINFO);
    pinstance->docinfo.lpszDocName = strclone(job_name);
    
    StartDoc(pinstance->printdlg.hDC, &pinstance->docinfo);
    StartPage(pinstance->printdlg.hDC);
    
    return SP_TRUE;
}

static spBool spEndDrawingPrinter(void *instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    EndPage(pinstance->printdlg.hDC);
    EndDoc(pinstance->printdlg.hDC);
    xfree(pinstance->docinfo.lpszDocName);
    
    return SP_TRUE;
}

static spBool spForwardPagePrinter(void *instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    EndPage(pinstance->printdlg.hDC);
    StartPage(pinstance->printdlg.hDC);
    
    return SP_TRUE;
}

static spBool spCopyImagePrinter(void *instance, void *src_image_arch,
				 int src_x, int src_y, int src_width, int src_height,
				 int dest_x, int dest_y, int dest_width, int dest_height)
{
    HDC memprdc;
    HBITMAP hbitmap;
    HGDIOBJ hold;
    spBool flag;
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    flag = SP_FALSE;
    if ((memprdc = CreateCompatibleDC(pinstance->printdlg.hDC)) != NULL) {
	if ((hbitmap = CreateCompatibleBitmap(pinstance->printdlg.hDC, src_width, src_height)) != NULL) {
	    hold = SelectObject(memprdc, hbitmap);

	    BitBlt(memprdc, 0, 0, src_width, src_height,
		   (HDC)src_image_arch, src_x, src_y, SRCCOPY);

	    if (GetDeviceCaps(pinstance->printdlg.hDC, RASTERCAPS) & RC_STRETCHBLT) {
		StretchBlt(pinstance->printdlg.hDC, dest_x, dest_y, dest_width, dest_height,
			   memprdc, 0, 0, src_width, src_height, SRCCOPY);
	    } else {
		BitBlt(pinstance->printdlg.hDC, dest_x, dest_y, src_width, src_height, 
		       memprdc, 0, 0, SRCCOPY);
	    }

	    SelectObject(memprdc, hold);
	    DeleteObject(hbitmap);
	    flag = SP_TRUE;
	}
	DeleteDC(memprdc);
    }
    
    return flag;
}

void spFillRectanglePrinter(void *instance, spGraphics graphics,
			    int x, int y, int width, int height)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    spDrawRectangleWin(pinstance->printdlg.hDC, graphics, SP_GRAPHICS_BRUSH,
		       g_null_pen, g_null_brush, x, y, width, height);

    return;
}

void spDrawRectanglePrinter(void *instance, spGraphics graphics,
			    int x, int y, int width, int height)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    spDrawRectangleWin(pinstance->printdlg.hDC, graphics, SP_GRAPHICS_PEN,
		       g_null_pen, g_null_brush, x, y, width, height);

    return;
}

void spFillArcPrinter(void *instance, spGraphics graphics,
		      int x, int y, int width, int height, double angle1, double angle2)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    spDrawArcWin(pinstance->printdlg.hDC, graphics, SP_GRAPHICS_BRUSH,
		 g_null_pen, g_null_brush,
		 x, y, width, height, angle1, angle2);
    
    return;
}
     
void spDrawArcPrinter(void *instance, spGraphics graphics,
		      int x, int y, int width, int height, double angle1, double angle2)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    spDrawArcWin(pinstance->printdlg.hDC, graphics, SP_GRAPHICS_PEN,
		 g_null_pen, g_null_brush,
		 x, y, width, height, angle1, angle2);

    return;
}

void spDrawLinePrinter(void *instance, spGraphics graphics,
		       int x1, int y1, int x2, int y2)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    spDrawLineWin(pinstance->printdlg.hDC, graphics, x1, y1, x2, y2);

    return;
}

void spDrawPointPrinter(void *instance, spGraphics graphics,
			int x, int y)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;

    spDrawPointWin(pinstance->printdlg.hDC, graphics, x, y);

    return;
}

void spDrawStringPrinter(void *instance, spGraphics graphics,
			 int x, int y, const char *string)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    spDrawStringWin(pinstance->printdlg.hDC, graphics, x, y, string);
}

spBool spGetStringExtentPrinter(void *instance, spGraphics graphics, const char *string,
				int *x, int *y, int *width, int *height, int *next_x)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    return spGetStringExtentWin(pinstance->printdlg.hDC, graphics, string, x, y, width, height, next_x);
}

spBool spFillPolygonPrinter(void *instance, spGraphics graphics, void *polygon_instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    
    return spDrawPolygonWin(pinstance->printdlg.hDC, graphics, SP_GRAPHICS_BRUSH,
			    g_null_pen, g_null_brush, (spPolygon)polygon_instance);
}

spBool spDrawPolygonPrinter(void *instance, spGraphics graphics, void *polygon_instance)
{
    spPluginInstancePrinter pinstance = (spPluginInstancePrinter)instance;
    spPolygon polygon = (spPolygon)polygon_instance;
    
    return spDrawPolygonWin(pinstance->printdlg.hDC, graphics, SP_GRAPHICS_PEN,
			    g_null_pen, g_null_brush, (spPolygon)polygon_instance);
}

#if 1
static spDrawingPluginRec sp_drawing_plugin_printer = {
    NULL,
    NULL,

    SP_PLUGIN_DRAWING,
    SP_DRAWING_PLUGIN_PRINTER,
    1,
    SP_PLUGIN_PRIORITY_MIDDLE,
    SP_DRAWING_PLUGIN_CAPS_MULTIPLE_PAGES,		/* caps */
    spInitPrinter,
    spFreePrinter,
    "Printer",
    "Printer Plugin for Windows  Version 0.1\nCopyright(C) Hideki Banno",
    
    spInitInstancePrinter,
    spFreeInstancePrinter,
    NULL,
    NULL,

    SP_DRAWING_DEVICE_PRINTER,

    NULL,
    NULL,
    NULL,
    NULL,

    NULL,
    spCreateComponentPrinter,
    spDestroyComponentPrinter,
    NULL/*spSetParamsPrinter*/,
    NULL/*spGetParamsPrinter*/,

    spOpenSetupDialogPrinter,
    
    spSetPageInfoPrinter,
    spGetPageInfoPrinter,
    spSetPageRangePrinter,
    spGetPageRangePrinter,
    NULL,

    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    
    NULL,
    spOpenTargetDialogPrinter,

    spCloseTargetPrinter,
    
    spIsDrawablePrinter,

    spGetDrawingCapsPrinter,
    spGetGraphicsModeCapsPrinter,
    NULL,
    spGetClientSizePrinter,
    spGetPhysicalSizePrinter,
    spSetScalingFactorPrinter,

    spStartDrawingPrinter,
    spEndDrawingPrinter,
    spForwardPagePrinter,

    NULL,
    spCopyImagePrinter,

    spFillRectanglePrinter,
    spDrawRectanglePrinter,
    spFillArcPrinter,
    spDrawArcPrinter,
    spDrawLinePrinter,
    spDrawPointPrinter,
    spDrawStringPrinter,
    spGetStringExtentPrinter,

    NULL,
    NULL,
    NULL,
    
    NULL,
    spFillPolygonPrinter,
    spDrawPolygonPrinter,
};

spBool spAddPrinterPlugin(void)
{
    return spAddPluginRecList((spPluginRec *)&sp_drawing_plugin_printer, SP_DRAWING_PLUGIN_PRINTER);
}

spBool spRemovePrinterPlugin(void)
{
    return spRemovePluginRecList((spPluginRec *)&sp_drawing_plugin_printer);
}
#endif
